@@ -85,16 +85,38 @@ In some situations, you may need to use `collector` to collect data back from ch
     pp.join()
 
     print(processed)
+
+You can count the executions in `collector` to estimate the progress. To get the progress of mapper, \
+create a progress function and set it in `ParallelProcessor`::
+
+    def progress(p):
+
+        # print('Total task: {}, Added to queue: {}, Mapper Loaded: {}, Mapper Processed {}'.format(
+        #    p['total'], p['added'], p['loaded'], p['processed']))
+        if p['processed'] % 10 == 0:
+            print('Progress: {}%'.format(100.0 * p['processed'] / p['total']))
+
+    pp = ParallelProcessor(8, mapper=mapper, progress=progress, progress_total=len(tasks))
+    pp.start()
+
+    for t in tasks:
+        pp.add_task(t)
+
 """
 
 import multiprocess as mp
 import threading
 import queue
 import inspect
+import sys
+import typing
 from typing import Callable, Iterable
 
 from pyrallel import Paralleller
 
+if sys.version_info >= (3, 8):
+    from pyrallel import ShmQueue
+
 
 class Mapper(object):
     """
@@ -161,10 +183,16 @@ class ProgressThread(threading.Thread):
     P_ADDED = 0
     P_LOADED = 1
     P_PROCESSED = 2
+    P_TOTAL = 3
 
-    def __init__(self, instance, progress, num_of_processor):
+    def __init__(self, instance, progress, progress_total, num_of_processor):
         super(ProgressThread, self).__init__()
-        self.progress_info = {ProgressThread.P_ADDED: 0, ProgressThread.P_LOADED: 0, ProgressThread.P_PROCESSED: 0}
+        self.progress_info = {
+            ProgressThread.P_ADDED: 0,
+            ProgressThread.P_LOADED: 0,
+            ProgressThread.P_PROCESSED: 0,
+            ProgressThread.P_TOTAL: progress_total
+        }
         self.mapper_progress_info = [ProgressThread.init_mapper_progress_info() for _ in range(num_of_processor)]
         self.instance = instance
         self.progress = progress
@@ -183,7 +211,13 @@ class ProgressThread(threading.Thread):
         for idx, mapper_progress_info in self.instance.get_progress():
             self.mapper_progress_info[idx] = mapper_progress_info
             self.refresh_progress_info()
-            self.progress(self.progress_info)
+            progress_info = {
+                'added': self.progress_info[ProgressThread.P_ADDED],
+                'loaded': self.progress_info[ProgressThread.P_LOADED],
+                'processed': self.progress_info[ProgressThread.P_PROCESSED],
+                'total': self.progress_info[ProgressThread.P_TOTAL],
+            }
+            self.progress(progress_info)
 
 
 class ParallelProcessor(Paralleller):
@@ -204,13 +238,29 @@ class ParallelProcessor(Paralleller):
                                 passed to `mapper` function. This has no effect for `Mapper` class.
                                 It defaults to False.
         batch_size (int, optional): Batch size, defaults to 1.
-        progress (Callable, optional): Progress inspection. Defaults to None.
-
+        progress (Callable, optional): Progress function, which takes a dictionary as input.
+                                The dictionary contains following keys: `total` can be set by `progress_total`,
+                                `added` indicates the number of tasks has been added to the queue,
+                                `loaded` indicates the number of tasks has been loaded to worker processes,
+                                `processed` indicates the number of tasks has been processed by worker processes.
+                                Defaults to None.
+        progress_total (int, optional): Total number of tasks. Defaults to None.
+        use_shm (bool, optional): When True, and when running on Python version 3.8 or later,
+                                use ShmQueue for higher performance.  Defaults to False.
+        enable_collector_queues (bool, optional): When True, create a collector queue for each
+                                processor.  When False, do not allocate collector queues, saving
+                                resources.  Defaults to True.
+        single_mapper_queue (bool, optional): When True, allocate a single mapper queue that will
+                                be shared between the worker processes.  Sending processes can
+                                go to sleep when the mapper queue is full.  When False, each process
+                                gets its own mapper queue, and CPU-intensive polling may be needed to
+                                find a mapper queue which can accept a new request.
 
     Note:
         - Do NOT implement heavy compute-intensive operations in collector, they should be in mapper.
         - Tune the value for queue size and batch size will optimize performance a lot.
         - `collector` only collects returns from `mapper` or `Mapper.process`.
+        - The frequency of executing `progress` function depends on CPU.
     """
 
     # Command format in queue. Represent in tuple.
@@ -219,15 +269,72 @@ class ParallelProcessor(Paralleller):
     CMD_DATA = 0
     CMD_STOP = 1
 
+    QSTATS_ON = 0
+    QSTATS_OFF = 1
+
     def __init__(self, num_of_processor: int, mapper: Callable, max_size_per_mapper_queue: int = 0,
                  collector: Callable = None, max_size_per_collector_queue: int = 0,
-                 enable_process_id: bool = False, batch_size: int = 1, progress=None):
+                 enable_process_id: bool = False, batch_size: int = 1, progress=None, progress_total = None,
+                 use_shm=False, enable_collector_queues=True,
+                 single_mapper_queue: bool = False):
         self.num_of_processor = num_of_processor
-        self.mapper_queues = [mp.Queue(maxsize=max_size_per_mapper_queue) for _ in range(num_of_processor)]
-        self.collector_queues = [mp.Queue(maxsize=max_size_per_collector_queue) for _ in range(num_of_processor)]
-        self.processes = [mp.Process(target=self._run, args=(i, self.mapper_queues[i], self.collector_queues[i]))
-                          for i in range(num_of_processor)]
-        self.progress_queues = [mp.Queue(maxsize=1) for _ in range(num_of_processor)]
+        self.single_mapper_queue = single_mapper_queue
+        if sys.version_info >= (3, 8):
+            self.collector_queues: typing.Optional[typing.Union[ShmQueue, mp.Queue]]
+        else:
+            self.collector_queues: typing.Optional[mp.Queue]
+        if use_shm:
+            if sys.version_info >= (3, 8):
+                if single_mapper_queue:
+                    self.mapper_queues = [ShmQueue(maxsize=max_size_per_mapper_queue * num_of_processor)]
+                else:
+                    self.mapper_queues = [ShmQueue(maxsize=max_size_per_mapper_queue) for _ in range(num_of_processor)]
+                if enable_collector_queues:
+                    self.collector_queues = [ShmQueue(maxsize=max_size_per_collector_queue) for _ in range(num_of_processor)]
+                else:
+                    self.collector_queues = None
+            else:
+                raise ValueError("shm not available in this version of Python.")
+        else:
+            if single_mapper_queue:
+                self.mapper_queues = [mp.Queue(maxsize=max_size_per_mapper_queue * num_of_processor)]
+            else:
+                self.mapper_queues = [mp.Queue(maxsize=max_size_per_mapper_queue) for _ in range(num_of_processor)]
+            if enable_collector_queues:
+                self.collector_queues = [mp.Queue(maxsize=max_size_per_collector_queue) for _ in range(num_of_processor)]
+                self.collector_qstats = [self.QSTATS_ON for _ in range(num_of_processor)]
+            else:
+                self.collector_queues = None
+                
+        if self.collector_queues is not None:
+            if single_mapper_queue:
+                self.processes = [mp.Process(target=self._run, args=(i, self.mapper_queues[0], self.collector_queues[i]))
+                                  for i in range(num_of_processor)]
+            else:
+                self.processes = [mp.Process(target=self._run, args=(i, self.mapper_queues[i], self.collector_queues[i]))
+                                  for i in range(num_of_processor)]
+        else:
+            if single_mapper_queue:
+                self.processes = [mp.Process(target=self._run, args=(i, self.mapper_queues[0], None))
+                                  for i in range(num_of_processor)]
+            else:
+                self.processes = [mp.Process(target=self._run, args=(i, self.mapper_queues[i], None))
+                                  for i in range(num_of_processor)]
+        if progress is not None:
+            if sys.version_info >= (3, 8):
+                self.progress_queues: typing.Optional[typing.Union[ShmQueue, mp.Queue]]
+            else:
+                self.progress_queues: typing.Optional[mp.Queue]
+            if use_shm:
+                if sys.version_info >= (3, 8):
+                    self.progress_queues = [ShmQueue(maxsize=1) for _ in range(num_of_processor)]
+                else:
+                    raise ValueError("shm not available in this version of Python.")
+            else:
+                self.progress_queues = [mp.Queue(maxsize=1) for _ in range(num_of_processor)]
+            self.progress_qstats = [self.QSTATS_ON for _ in range(num_of_processor)]
+        else:
+            self.progress_queues = None
         self.progress = progress
 
         ctx = self
@@ -243,7 +350,6 @@ class ParallelProcessor(Paralleller):
 
         self.collector = collector
         self.mapper_queue_index = 0
-        self.collector_queue_index = 0
         self.enable_process_id = enable_process_id
         self.batch_size = batch_size
         self.batch_data = []
@@ -255,7 +361,7 @@ class ParallelProcessor(Paralleller):
             self.collector_thread = CollectorThread(self, collector)
 
         if progress:
-            self.progress_thread = ProgressThread(self, progress, num_of_processor)
+            self.progress_thread = ProgressThread(self, progress, progress_total, num_of_processor)
 
     def start(self):
         """
@@ -278,6 +384,18 @@ class ParallelProcessor(Paralleller):
             self.progress_thread.join()
         for p in self.processes:
             p.join()
+        for q in self.mapper_queues:
+            q.close()
+            q.join_thread()
+        if self.collector_queues is not None:
+            for q in self.collector_queues:
+                q.close()
+                q.join_thread()
+        if self.progress_queues is not None:
+            for q in self.progress_queues:
+                q.close()
+                q.join_thread()
+                pass
 
     def task_done(self):
         """
@@ -288,13 +406,20 @@ class ParallelProcessor(Paralleller):
             self._add_task(self.batch_data)
             self.batch_data = []
 
-        for q in self.mapper_queues:
-            q.put((ParallelProcessor.CMD_STOP,))
+        for i in range(self.num_of_processor):
+            if self.single_mapper_queue:
+                self.mapper_queues[0].put((ParallelProcessor.CMD_STOP,))
+            else:
+                self.mapper_queues[i].put((ParallelProcessor.CMD_STOP,))
 
     def add_task(self, *args, **kwargs):
         """
-        Add data to one of the mapper queues.
-        (main process, unblocked, using round robin to find next available queue)
+        Add data to one a mapper queue.
+
+        When a single mapper queue is in use, put the process to sleep if the
+        queue is full.  When multiple mapper queues are in use (one per process),
+        use CPU-intensive polling (round-robin processing) to find the next available
+        queue. (main process, blocked or unblocked depending upon single_mapper_queue)
         """
         self.batch_data.append((args, kwargs))
         if self.progress:
@@ -305,18 +430,21 @@ class ParallelProcessor(Paralleller):
             self.batch_data = []  # reset buffer
 
     def _add_task(self, batched_args):
-        while True:
-            q = self.mapper_queues[self.mapper_queue_index]
-            self.mapper_queue_index = (self.mapper_queue_index + 1) % self.num_of_processor
-            try:
-                q.put_nowait((ParallelProcessor.CMD_DATA, batched_args))
-                return  # put in
-            except queue.Full:
-                continue  # find next available
+        if self.single_mapper_queue:
+            self.mapper_queues[0].put((ParallelProcessor.CMD_DATA, batched_args))
+        else:
+            while True:
+                q = self.mapper_queues[self.mapper_queue_index]
+                self.mapper_queue_index = (self.mapper_queue_index + 1) % self.num_of_processor
+                try:
+                    q.put_nowait((ParallelProcessor.CMD_DATA, batched_args))
+                    return  # put in
+                except queue.Full:
+                    continue  # find next available
 
-    def _run(self, idx: int, mapper_queue: mp.Queue, collector_queue: mp.Queue):
+    def _run(self, idx: int, mapper_queue: mp.Queue, collector_queue: typing.Optional[mp.Queue]):
         """
-        Process’s activity. It handles queue IO and invokes user's mapper handler.
+        Process's activity. It handles queue IO and invokes user's mapper handler.
         (subprocess, blocked, only two queues can be used to communicate with main process)
         """
         with self.mapper(idx) as mapper:
@@ -325,7 +453,7 @@ class ParallelProcessor(Paralleller):
                 if data[0] == ParallelProcessor.CMD_STOP:
                     # print(idx, 'stop')
                     self._update_progress(mapper, finish=True)
-                    if self.collector:
+                    if self.collector and collector_queue is not None:
                         collector_queue.put((ParallelProcessor.CMD_STOP,))
                     return
                 elif data[0] == ParallelProcessor.CMD_DATA:
@@ -336,11 +464,12 @@ class ParallelProcessor(Paralleller):
                         self._update_progress(mapper, type_=ProgressThread.P_LOADED)
                         result = mapper.process(*args, **kwargs)
                         self._update_progress(mapper, type_=ProgressThread.P_PROCESSED)
-                        if self.collector:
-                            if not isinstance(result, tuple):  # collector must represent as tuple
-                                result = (result,)
+                        if collector_queue is not None:
+                            if self.collector:
+                                if not isinstance(result, tuple):  # collector must represent as tuple
+                                    result = (result,)
                             batch_result.append(result)
-                    if len(batch_result) > 0:
+                    if collector_queue is not None and len(batch_result) > 0:
                         collector_queue.put((ParallelProcessor.CMD_DATA, batch_result))
                         batch_result = []  # reset buffer
 
@@ -365,49 +494,52 @@ class ParallelProcessor(Paralleller):
         """
         if not self.collector:
             return
+
+        idx = 0
         while True:
-            # print(self.collector_queues)
-            q = self.collector_queues[self.collector_queue_index]
+            # all queues finished
+            if sum([int(s == self.QSTATS_OFF) for s in self.collector_qstats]) == self.num_of_processor:
+                return
+
+            # get next unfinished queue
+            while self.collector_qstats[idx] == self.QSTATS_OFF:
+                idx = (idx + 1) % self.num_of_processor
+            q = self.collector_queues[idx]
+
             try:
                 data = q.get_nowait()  # get out
                 if data[0] == ParallelProcessor.CMD_STOP:
-                    del self.collector_queues[self.collector_queue_index]  # remove queue if it's finished
+                    self.collector_qstats[idx] = self.QSTATS_OFF
                 elif data[0] == ParallelProcessor.CMD_DATA:
                     yield data[1]
             except queue.Empty:
                 continue  # find next available
             finally:
-                if len(self.collector_queues) == 0:  # all finished
-                    return
-                self.collector_queue_index = (self.collector_queue_index + 1) % len(self.collector_queues)
+                idx = (idx + 1) % self.num_of_processor
 
     def get_progress(self):
         """
-        Get progress infomation from each mapper.
+        Get progress information from each mapper.
         (main process)
         """
         if not self.progress:
             return
 
         idx = 0
-        finish_num = 0
         while True:
-            if finish_num == self.num_of_processor:
+            # all queues finished
+            if sum([int(s == self.QSTATS_OFF) for s in self.progress_qstats]) == self.num_of_processor:
                 return
 
-            # get next not-None queue
-            q = None
-            tmp_idx = idx
-            while not q:
-                q = self.progress_queues[tmp_idx]
-                idx = tmp_idx
-                tmp_idx = (tmp_idx + 1) % self.num_of_processor
+            # get next unfinished queue
+            while self.progress_qstats[idx] == self.QSTATS_OFF:
+                idx = (idx + 1) % self.num_of_processor
+            q = self.progress_queues[idx]
 
             try:
                 data = q.get_nowait()
                 if data[0] == ParallelProcessor.CMD_STOP:
-                    self.progress_queues[idx] = None  # set to None if it's finished
-                    finish_num += 1
+                    self.progress_qstats[idx] = self.QSTATS_OFF
                 elif data[0] == ParallelProcessor.CMD_DATA:
                     pass
                 yield idx, data[1]
